/**
 * INDOORS Data Acquisition Platform
 * Core backend module
 * Run it with Node.js
 * 
 * (c) 2015-2017 Andrei Popleteev, INDOORS project
 */
var express = require('express');
var bodyParser = require('body-parser'); // for parsing POST requests
var app = express();
app.use(express.static('ui'));
app.use(bodyParser.json());       // to support JSON-encoded bodies
app.use(bodyParser.urlencoded({extended: true})); // to support URL-encoded bodies
var http = require('http').Server(app);
var io = require('socket.io')(http);
var spawn = require('child_process').spawn;
var fs = require('fs');
var mkdirp = require('mkdirp');

// Root directory to store datasets
//const SESSION_ROOT_DIR = "/tmp/indoors-data"; // no trailing slash!
const SESSION_ROOT_DIR = "/Users/Shared/indoors-data/new/raw"; // no trailing slash!
var sessionId = undefined;
var sessionDir = undefined; // dataset dir for the current session (WITHOUT trailing slash)
var sessionParams = undefined; // set on start of new session

// Acquisition parameters: duration, frequencies
const SAMPLE_DURATION_SECONDS = 5;
var rfConfig = require('./rf-config.js');

// Child process responsible for sampling
var samplingProcess = undefined;

// Sends a string to the client via Socket.io
function pushMessage(msg) {
  io.emit('child message', msg);
}

/**
 * Since RF acqusition script logs a lot of redundant stuff,
 * we filter it out at the server and never send to the client's log.
 */
const IGNORED_MESSAGES = [
  "Win32; Microsoft Visual C++ version 12.0; Boost_105500; UHD_003.008.002-release",
  "Mac OS; Clang version 6.1.0 (clang-602.0.49); Boost_105800; UHD_003.008.004-MacPorts-Release",
  "Mac OS; Clang version 6.1.0 (clang-602.0.49); Boost_105900; UHD_003.009.001-MacPorts-Release",
];
function isIgnoredMessage(msg) {
  for (var i in IGNORED_MESSAGES) {
    if (msg === IGNORED_MESSAGES[i])
      return true;
  }
  return false;
}

/**
 * Acquires and saves the given RF bands for the given point.
 * Returns asynchronously; calls doneCallback once done.
 * @param sourceBands   array with band names to acquire
 * @param pointParams   sampling location params
 * @param doneCallback  called once the sampling is over (or threw an error)
 * @return process ID of the RF sampling script
 */
function acquireRFSample(sourceBands, pointParams, doneCallback) {
  if (!!samplingProcess) {
    console.log("WTF: sampling process still running");
    pushMessage("Internal error: sampling process still running");
    return;
  }

  const SEP = "_";  // separator string for file name components
  const SUBSEP = "~"; // separator string to use inside components

  var pointDir = sessionDir + "/" + 
      // Each point has WGS84 coords
      "WGS" + SUBSEP + parseFloat(pointParams.coords[0]).toFixed(6) + SUBSEP + parseFloat(pointParams.coords[1]).toFixed(6) + SEP +
      // _Some_ points have CRS coords
      (!pointParams.crs ? "" : "CRS" + SUBSEP + pointParams.crs.join(SUBSEP) + SEP) +
      // User-defined antenna direction
      pointParams.direction; 

  // Now build the command-line params for the sampling utility
  var cliScriptParams = ["--outdir", pointDir];

  for (var iBandId in sourceBands) {
    var bandId = sourceBands[iBandId];
    var duration = rfConfig.getDuration(bandId);
    var gain = rfConfig.getGain(bandId);
    var bandParams = rfConfig.getBandParams(bandId);
    for (var iChannel in bandParams) {
      var channelParams = bandParams[iChannel];
      var antennaParams = rfConfig.getAntenna(channelParams.antenna);
      var sampleSpec = 
            channelParams.id + ',' + 
            channelParams.centerFreq / 1e6 + ',' +
            channelParams.rate / 1e6 + ',' +
            channelParams.bandwidth / 1e6 + ',' +
            antennaParams.subdev + ',' + 
            antennaParams.ant + ',' +
            gain + ',' + 
            duration;
      //console.log("Sample spec: " + sampleSpec);
      cliScriptParams.push("--sample");
      cliScriptParams.push(sampleSpec);
    }
  }

  console.log("Acquiring " + pointDir); // to server log
  pushMessage("Acquiring: " + pointDir); // to client's log

  samplingProcess = spawn('sdr/radio_sampler', cliScriptParams);

  samplingProcess.on('error', function(code) {
      console.log('error: ' + code);
  	  io.emit('sampling error', 'Sampling error (code ' + code + ').');
  });
  samplingProcess.stdout.on('data', function(data) {
    // Forwards SDR messages to the UI
    var msg = data.toString().trim();
    console.log('stdout: ' + msg);
    if (!isIgnoredMessage(msg))
      pushMessage(msg);
  });
  samplingProcess.stderr.on('data', function(data) {
    var msg = data.toString().trim();
    console.log('stderr: ' + msg);
    pushMessage(msg);
  });
  samplingProcess.on('exit', function(code) {
    samplingProcess = undefined;
    // originally code is unsigned, which looks weird for negative (error) return codes
    code = (code >> 0); // this makes it signed
    console.log('RadioSampler returned ' + code);
    /*
      Possible sampling script exit codes:
        EXIT_FAILED_TO_CREATE_OUTPUT_DIR = -2
        EXIT_INCORRECT_PARAMETERS        = -1
        EXIT_SUCCESS                     = 0
        EXIT_USER_INTERRUPT              = 1
    */
    if (code == 0) {
      pushMessage("Sample ready.");
      if (doneCallback)
        doneCallback();
    } else if (code == 1) { // interrupted by request
      // stopped by the interruption signal we've sent before
      io.emit('sampling cancelled');
    } else {
      io.emit('sampling error', 'Sampling error (code ' + code + '). Any remaining bands were ignored. Fix the issue and reacqure the sample.');
    }
  });

//  console.log('Spawned child pid: ' + samplingProcess.pid);
  samplingProcess.stdin.end();
  return samplingProcess.pid;
}

/** Sends the UI page to the client */
app.get('/', function(req, res) {
  console.log("Got INDEX request for " + __dirname + '/ui/index.html');
  res.sendFile(__dirname + '/ui/index.html');
});

/** Called when the UI wants to start an acquisition session */
app.post('/startSession', function(req, res) {
  console.log("Got startSession request");
  console.log(JSON.stringify(req.body));
  sessionParams = req.body;

  var pointSetId = sessionParams.pointSetId;
  var pointSetParts = sessionParams.pointSetParts;
  var timestamp = (new Date()).toISOString()
      // convert format to YYYYMMDD-hhmmss
      .replace(/[-\:]/ig, "").replace("T","-").replace(/\..+Z$/ig, "");
  sessionId = pointSetId.join("-") + "_" + timestamp;
  sessionDir = SESSION_ROOT_DIR + "/" + pointSetParts.join('/') + "/" + timestamp;

  // ensure the directory exists
  mkdirp(sessionDir, function (err) {
    if (err) {
      var msg = "Cannot create session directory (" + sessionDir + "): " + err;
      console.error(msg);
      pushMessage(msg);
      sessionId = undefined;
      sessionDir = undefined;
      sessionParams = undefined;
      res.status(500).end('Failed to create session dir.');
    } else {
      res.send(JSON.stringify({'sessionId': sessionId}));
      res.status(200).end('OK');

      // Integrate the general RF configuration for the record
      sessionParams.rfConfig = rfConfig.getConfig();

      // Now save the session info file
      var sessionInfo = JSON.stringify(sessionParams, null, 2);
      fs.writeFile(sessionDir + "/session.info", sessionInfo, function(err) {
        if(err) {
          pushMessage("Error saving session file: " + err);
          return console.log(err);
        }
        console.log("The file was saved!");
      }); 
    }
  });
});

/** Called when the UI wants to start sampling a test location */
app.post('/startSampling', function(req, res) {
  console.log("Got startSampling request");
  if (!sessionId) {
    res.status(500).end("Can acquire samples only when a session is started.")
    return;
  }

  var pointParams = req.body.pointParams;
  // Acquire samples for the requested bands and point
  acquireRFSample(sessionParams.bands, pointParams, function() {
    // completion callback
    io.emit('sample ready');
  });

  // Sampling can take a lot of time, so return immediately to avoid HTTP timeout.
  // Progress will be reported via socket.io
  res.status(200).end('OK');
});

/** Called when the UI wants to cancel sampling in progress */
app.post('/stopSampling', function(req, res) {
  console.log("Got stopSampling request");
  if (!samplingProcess) {
    res.status(400).end("Cannot stop sampling, it is not running.");
    return;
  }

  // This does not kill the process directly, but only sends it a signal 
  // to stop sampling, close the device and quit.
  samplingProcess.kill('SIGINT'); 
  res.status(200).end('OK');
});

io.on('connection', function(socket) {
  console.log('user connected');

  socket.on('message', function(msg) {
    console.log('message from client: ' + msg);
  });
  socket.on('disconnect', function() {
    console.log('user disconnected');
  });
});

http.listen(8080);
