//
// Copyright 2010-2011,2014 Ettus Research LLC
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/*
  Modified by Andrei Popleteev.
  Copyright (c) 2015-2017 INDOORS Project.
*/


#include <uhd/types/tune_request.hpp>
#include <uhd/utils/thread_priority.hpp>
#include <uhd/utils/safe_main.hpp>
#include <uhd/usrp/multi_usrp.hpp>
#include <uhd/exception.hpp>
#include <boost/program_options.hpp>
#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
//#include <boost/asio.hpp>
//#include <boost/asio/posix/stream_descriptor.hpp>
#include <iostream>
#include <fstream>
#include <csignal>
#include <complex>

namespace po = boost::program_options;
namespace fs = boost::filesystem;

#define WIRE_FORMAT "sc16"
#define CPU_FORMAT "sc16"
#define SETUP_TIME_SECONDS 1.0
// If there was an overflow/timeout/other SDR error, retry sampling up to this number of times
#define RETRY_ATTEMPTS 3

// Maximum allowed master clock rate. Up to 61.44 MHz is possible, but not recommended.
// http://files.ettus.com/manual/page_usrp_b200.html
#define MAX_MASTER_CLOCK 56e6

enum ExitCodes {
    EXIT_FAILED_TO_CREATE_OUTPUT_DIR = -2,
    EXIT_INCORRECT_PARAMETERS        = -1,
    // EXIT_SUCCESS = 0   already defined in stdlib.h
    EXIT_USER_INTERRUPT              = 1
};

static bool stop_signal_called = false;
void sig_int_handler(int){stop_signal_called = true;}

typedef struct  {
    std::string channel_id; // Human-readable RF channel ID, e.g. "fm" or "wifi24_06"
    double center_freq;
    double rate;
    double bandwidth;
    std::string subdev;  // e.g. "A:A" or "A:B"
    std::string antenna; // e.g. "RX2" or "TX/RX"
    double gain;         // in dB
    double duration;     // in seconds
} sample_config;

/* Custom handler of UHD console messages */
void uhd_msg_handler(uhd::msg::type_t type, const std::string& msg){
    //do nothing, keep it silent
}

/* 
 * Stores samples to file. 
 * Returns an error code or zero if successful. 
 */
int recv_to_file(
               uhd::usrp::multi_usrp::sptr usrp,
               const std::string &file,
               unsigned long long num_requested_samples,
               bool bw_summary = false,
               bool stats = false
               ){
    int result = 0;
    size_t samps_per_buff = size_t(num_requested_samples);
    unsigned long long num_total_samps = 0;
    //create a receive streamer
    uhd::stream_args_t stream_args(CPU_FORMAT, WIRE_FORMAT);
    uhd::rx_streamer::sptr rx_stream = usrp->get_rx_stream(stream_args);
    
    uhd::rx_metadata_t md;
    std::vector<std::complex<short>> buff(samps_per_buff);
    std::ofstream outfile;
    if (not file.empty())
        outfile.open(file.c_str(), std::ofstream::binary);

    
    //setup streaming
    uhd::stream_cmd_t stream_cmd((num_requested_samples == 0)?
                                 uhd::stream_cmd_t::STREAM_MODE_START_CONTINUOUS:
                                 uhd::stream_cmd_t::STREAM_MODE_NUM_SAMPS_AND_DONE
                                 );
    stream_cmd.num_samps = size_t(num_requested_samples);
    stream_cmd.stream_now = true;
    stream_cmd.time_spec = uhd::time_spec_t();
    rx_stream->issue_stream_cmd(stream_cmd);
    
    boost::system_time now = boost::get_system_time();
    boost::system_time start = now;
    boost::posix_time::time_duration ticks_diff;

    while(not stop_signal_called and (num_requested_samples != num_total_samps or num_requested_samples == 0)) {
        //TODO: 3.0 below is the timeout -- it must be sufficient for the requested number of samples!
        size_t num_rx_samps = rx_stream->recv(&buff.front(), buff.size(), md, 3.0, false);
        now = boost::get_system_time();
        
        if (md.error_code == uhd::rx_metadata_t::ERROR_CODE_TIMEOUT) {
            std::cerr << boost::format("Timeout while streaming") << std::endl;
            result = 1;
            break;
        }
        if (md.error_code == uhd::rx_metadata_t::ERROR_CODE_OVERFLOW){
            std::cerr << "Overflow" << std::endl;
            result = 2;
            break;
        }
        if (md.error_code != uhd::rx_metadata_t::ERROR_CODE_NONE){
            std::string error = str(boost::format("Receiver error: %s") % md.strerror());
            std::cerr << error << std::endl;
            result = 3;
            break;
        }
        
        num_total_samps += num_rx_samps;
        
        if (outfile.is_open())
            outfile.write((const char*)&buff.front(), num_rx_samps*sizeof(std::complex<short>));
    }
    
    stream_cmd.stream_mode = uhd::stream_cmd_t::STREAM_MODE_STOP_CONTINUOUS;
    rx_stream->issue_stream_cmd(stream_cmd);
    
    //Flush RX buffers by receiving until timeout
    //http://lists.ettus.com/pipermail/usrp-users_lists.ettus.com/2013-February/006117.
    // Reusing the same buffer
    size_t num_rx_samps_left = rx_stream->recv(&buff.front(), buff.size(), md, 0.1, false);
    std::cout << boost::format("Flushed %d samples after stopping.") % (num_rx_samps_left) << std::endl;
    
    if (outfile.is_open())
        outfile.close();
    
    if (stats) {
        ticks_diff = now - start;
        double t = (double)ticks_diff.ticks() / (double)boost::posix_time::time_duration::ticks_per_second();
        //std::cout << boost::format("Received %d samples in %f seconds") % num_total_samps % t << std::endl;
        double r = (double)num_total_samps / t;
        std::cout << boost::format("Got %.1fM samples at %.1f Msps") % (num_total_samps/1e6) % (r/1e6) << std::endl;
    }
    return result;
}

typedef boost::function<uhd::sensor_value_t (const std::string&)> get_sensor_fn_t;

/*
 * Waits for up to setup_time seconds for the local oscillator (LO) to settle.
 * Returns true if successful, false for timeout.
 */
bool check_local_oscillator_locked(uhd::usrp::multi_usrp::sptr usrp, double setup_time) {
    std::vector<std::string> sensor_names = usrp->get_rx_sensor_names();
    std::string lo_sensor_name = "lo_locked";
    if (std::find(sensor_names.begin(), sensor_names.end(), lo_sensor_name) == sensor_names.end())
        throw std::runtime_error(str(boost::format("Ooopsie... %s sensor is not found?!") % lo_sensor_name));
    
    bool is_locked = false;
    boost::system_time start = boost::get_system_time();
    while (true) {
        is_locked = usrp->get_rx_sensor(lo_sensor_name).to_bool();
        if (is_locked) {
            break;
        }
        if (boost::get_system_time() > (start + boost::posix_time::seconds(setup_time))) {
            std::cerr << "Timeout waiting for" << lo_sensor_name << std::endl;
            break;
        }
        std::cout << ".";
        boost::this_thread::sleep(boost::posix_time::milliseconds(100));
    }
    return is_locked;
}

/**
 * Parses command line arguments to sample specifications.
 */
void parse_sampling_params(const std::vector<std::string> args, std::vector<sample_config>& samples) {
    const int N_TOKENS = 8;
    BOOST_FOREACH(const std::string arg, args) {
        // The arg should contain exactly N_TOKENS comma-separated values:
        // channel_id,freq_mhz, rate_msps, bw_mhz, subdev, antenna, gain, duration_seconds
        std::vector<std::string> tokens;
        boost::algorithm::split(tokens, arg, boost::algorithm::is_any_of(","));
        if (tokens.size() < N_TOKENS) {
            throw std::runtime_error(str(boost::format("Too few tokens in \"%s\" (expected %d)") % arg % N_TOKENS));
        } else if (tokens.size() > N_TOKENS) {
            throw std::runtime_error(str(boost::format("Too many tokens in \"%s\" (expected %d)") % arg % N_TOKENS));
        }
        
        sample_config smp;
        try {
            smp.channel_id   = tokens[0];
            smp.center_freq  = 1e6 * boost::lexical_cast<double>(tokens[1]);
            smp.rate         = 1e6 * boost::lexical_cast<double>(tokens[2]);
            smp.bandwidth    = 1e6 * boost::lexical_cast<double>(tokens[3]);
            smp.subdev       = tokens[4];
            smp.antenna      = tokens[5];
            smp.gain         = boost::lexical_cast<double>(tokens[6]);
            smp.duration     = boost::lexical_cast<double>(tokens[7]);
        } catch(boost::bad_lexical_cast &) {
            throw std::runtime_error(str(boost::format("Cannot parse \"%s\"") % arg));
        }
        std::cout << str(
                boost::format("%s: Center %4.3f MHz, Rate %2.0f Msps, BW %2.0f MHz, Subdev: %s, Ant: %s, Gain %.0f dB. Duration %.1f s")
                            % smp.channel_id % (smp.center_freq/1e6) % (smp.rate/1e6) % (smp.bandwidth/1e6) % smp.subdev % smp.antenna % smp.gain % smp.duration
                ) << std::endl;
        samples.push_back(smp);
    }
}

/**
 * Configures the USRP to the given params and streams duration_seconds of data to a file.
 * Returns an error code or zero in case of success.
 */
int tune_and_save(
            uhd::usrp::multi_usrp::sptr usrp,
            const std::string &subdev,
            const std::string &ant,
            const double gain, 
            const double freq, 
            const double rate,
            const double bw,
            const double duration_seconds,
            const std::string &file_name
            ) {
    //always select the subdevice first, the channel mapping affects the other settings
    usrp->set_rx_subdev_spec(subdev);
    
    // As of 2015, the AGC is off by default, but there were a few commits where it was on.
    // http://lists.ettus.com/pipermail/usrp-users_lists.ettus.com/2015-August/015428.html
    // So we explicitly disable it just in case.
    usrp->set_rx_agc(false);
    
    
    // Automatic master clock rate selection does not seem to work in UHD_003.009.001
    // So we explicitly set it to the highest possible multiple of the sampling rate.
    usrp->set_master_clock_rate(rate * floor(MAX_MASTER_CLOCK / rate));
    usrp->set_rx_rate(rate);
    //std::cout << boost::format("Requested RX rate %.3f, got %.3f") % (rate/1e6) % (usrp->get_rx_rate()/1e6) << std::endl;
    
    uhd::tune_request_t tune_request(freq);
    usrp->set_rx_freq(tune_request);
    usrp->set_rx_gain(gain);
    usrp->set_rx_bandwidth(bw);
    usrp->set_rx_antenna(ant);

    std::cout << boost::format("Freq: %.1f MHz, %.1f Msps, BW %.1f MHz") % (usrp->get_rx_freq()/1e6) % (usrp->get_rx_rate()/1e6) % (usrp->get_rx_bandwidth()/1e6) << std::endl;
    std::cout << boost::format("Streaming %s %s (%.0f dB) to %s") % usrp->get_rx_antenna() % subdev % usrp->get_rx_gain() % file_name << std::endl;
    
    // make sure the RF front-end has settled
    if (not check_local_oscillator_locked(usrp, SETUP_TIME_SECONDS)) {
        return -1; // Timeout while waiting for LO lock
    }
    
//    //DEBUG
//    std::cout << boost::format("Actual RX gain: %.1f dB (normalized %.1f), master clock: %.3f MHz")
//            % usrp->get_rx_gain()
//            % (usrp->get_normalized_rx_gain())
//            % (usrp->get_master_clock_rate() / 1e6) << std::endl;

    //recv to file
    for (int retriesLeft = RETRY_ATTEMPTS; (retriesLeft >= 0) && !stop_signal_called; retriesLeft--) {
        int err = recv_to_file(usrp, file_name,
                               rate * duration_seconds, // number of samples to receive
                               false, // show progress while streaming - disabled
                               true); // show stats after receiving
        if (err == 0) // all good
            break;
        
        if (retriesLeft == 0) {
            // fatal error, cannot recover
            throw std::runtime_error(str(boost::format("*** ERROR: Fatal sampling error %d") % err));
        } else {
            std::cerr << str(boost::format("*** WARNING: Acquisition error %d, retrying (%d attempts left)") % err % retriesLeft) << std::endl;
            boost::this_thread::sleep(boost::posix_time::milliseconds(1000));
        }
    }
    
    std::cout << std::endl;
    return 0; // everything was OK
}

int UHD_SAFE_MAIN(int argc, char *argv[]){
    uhd::set_thread_priority_safe();
    
    std::cout << "INDOORS radio sampler" << std::endl;
    
    //variables to be set by po
    std::string args, output_path;
    std::vector<std::string> sample_args;
    
    //setup the program options
    po::options_description desc("Available options");
    desc.add_options()
    ("help", "help message")
    ("args", po::value<std::string>(&args)->default_value(""), "multi uhd device address args")
    ("outdir", po::value<std::string>(&output_path), "path to the output directory")
    ("null", "run without writing anything to files")
    // Multiple "-sample" args are ok.
    // Example: "-sample fm,98,20,20,A:B,RX2,40,5 -sample wifi24_00,2412,20,20,A:A,TX/RX,40,0.5"
    ("sample", po::value<std::vector<std::string>>(&sample_args), "Sample specification (channel_id,freq_mhz,rate_msps,bw_mhz,subdev,antenna,gain,duration_seconds)")
    ;
    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, desc), vm);
    po::notify(vm);
    
    //print the help message
    if (vm.count("help")) {
        std::cout
        << desc << std::endl
        << std::endl
        << "This application acquires and stores RF samples from a USRP device to files."
        << std::endl;
        return EXIT_SUCCESS;
    }

    if (not vm.count("sample")) {
        std::cout << "No --sample arg specified, nothing to do." << std::endl;
        return EXIT_INCORRECT_PARAMETERS;
    }

    std::vector<sample_config> samples;
    parse_sampling_params(sample_args, samples);

    boost::system::error_code err_code;
    fs::path target_dir = fs::absolute(fs::path(output_path));
    fs::create_directories(target_dir, err_code);
    if (err_code) {
        std::cerr << "Failed to create output directory: " << target_dir << std::endl;
        return EXIT_FAILED_TO_CREATE_OUTPUT_DIR;
    }

    bool skip_writing = vm.count("null") > 0;

    
    std::signal(SIGINT, &sig_int_handler);
//    for (int i = 0; i < 6; i++) {
//        std::cout << "Step" << i << std::endl;;
//        boost::this_thread::sleep(boost::posix_time::milliseconds(1000));
//        if (stop_signal_called) {
//            std::cout << "Sampling cancelled" << std::endl;
//            std::cout.flush();
//            break;
//        }
//    }
    
    //create a usrp device
    std::cout << std::endl;
    uhd::usrp::multi_usrp::sptr usrp = uhd::usrp::multi_usrp::make(args);
    std::cout << std::endl;
    
    //hide excessive messages from UHD
    uhd::msg::register_handler(uhd_msg_handler);
    
    //Lock mboard clocks
    usrp->set_clock_source("internal");

    boost::system_time start = boost::get_system_time();    
    BOOST_FOREACH(const sample_config& smp, samples) {
        fs::path file_path = target_dir / str(boost::format("%s_%.1f_%.0f.dat") % smp.channel_id % (smp.center_freq/1e6) % (smp.rate/1e6));

        tune_and_save(usrp, smp.subdev, smp.antenna, smp.gain, 
            smp.center_freq, smp.rate, smp.bandwidth, smp.duration, 
            skip_writing ? "" : file_path.string());
        if (stop_signal_called)
            break;
    }

    if (stop_signal_called) {
        std::cout << "Sampling cancelled" << std::endl;
        std::cout.flush();
        return EXIT_USER_INTERRUPT;
    } else {
        boost::posix_time::time_duration ticks_diff = boost::get_system_time() - start;
        double seconds = (double)ticks_diff.ticks() / (double)boost::posix_time::time_duration::ticks_per_second();
        std::cout << std::endl << boost::format("All done in %.0f seconds.") % seconds << std::endl;
        return EXIT_SUCCESS;
    }
}
