var IndoorsApp = function(config) {
	const SERVER_URL = "";
	const WEATHER_API_URL = "https://api.forecast.io/forecast/dd4591617697549d5f465a5aee2f17a6/";
	const FLOORMAP_PATH = "testbeds/"; // must include the trailing slash

	// Control elements
	var elHighlightOverlay = $("#highlightOverlay");
	var elSessionButton = $("#sessionButton");
	var elGetSampleButton = $("#getSampleButton");
	var elStopSamplingButton = $("#stopSamplingButton");
	var elPointSetSelector = $("#pointSetSelector");
	var elWeatherLoader = $("#weatherLoader");
	var elWeatherSummary = $("#weatherSummary");
	var elComments = $("#comments");
	var elCrowdActivity = $("#crowdActivity");
	var elAcquireFM = $("#acquireFM");
	var elAcquireDVB = $("#acquireDVB");
	var elAcquireGSM900down = $("#acquireGSM900down");
	var elAcquireGSM1800down = $("#acquireGSM1800down");
	var elAcquireWiFi24 = $("#acquireWiFi24");
	var elAcquireWiFi50 = $("#acquireWiFi50");

	this.config = config;
	var log = config.userLog; // function which adds lines to user-visible log area
    var playBeep = config.playBeep; // function which plays a short sound clip

	var sessionId = undefined; // current measurement session ID (string)
	var socket = io(); // socket.io channel
	var weather = {}; // current weather conditions
	var selectedPoint = null; // Params of the currently selected point/marker.
	var samplingInProgress = false;

	var leafletMap;
	var activeOverlays = {
		'mapLayer': undefined,
		'pointSetLayer': undefined,
	};
	var pointMarkerIcon = L.icon({
			iconUrl: '/js/images/marker32.png',
			iconSize: [16, 16],
			iconAnchor: [8, 8],
			popupAnchor: [0, -10]
		});

	var initMap = function(allTestbeds) {
		console.log("initMap()");
		leafletMap = L.map(config.mapElementName, {
			measureControl: true,
			autoZIndex: true,
			closePopupOnClick: false,
		});
		leafletMap.addLayer(L.tileLayer(
			'http://{s}.tile.osm.org/{z}/{x}/{y}.png', {
				maxZoom: 24,
				attribution: '&copy; <a href="http://osm.org/copyright">OpenStreetMap</a> contributors'
		}));
		/*
		// Uncomment below to use offline tiles
		leafletMap.addLayer(L.tileLayer('tiles/{z}/{x}/{y}.png', {
			maxZoom: 24,
			attribution: 'Map data &copy; <a href="http://openstreetmap.org">OpenStreetMap</a> contributors, <a href="http://creativecommons.org/licenses/by-sa/2.0/">CC-BY-SA</a>'
		}));
		*/
		console.log("Added base OSM layer");
		// Debug coordinate indicator
		leafletMap.on('click', function(e) {
			var popup = L.popup({
						"autoPan": true,
						"closeButton": false,
						"closeOnClick": true,
					});
			popup.setLatLng(e.latlng)
				.setContent(
						e.latlng.lat.toFixed(6) + ", "
						+ e.latlng.lng.toFixed(6))
				.openOn(leafletMap);
		});

		initPointSetSelector(allTestbeds);
	};

	var initPointSetSelector = function(allTestbeds) {
		// Fill pointSet selector
		for (var iTestbed in allTestbeds) {
			var testbed = allTestbeds[iTestbed];
			for (var iFloor in testbed.floors) {
				var floor = testbed.floors[iFloor];
				for (var iPointSet in floor.pointSets) {
					var pointSet = floor.pointSets[iPointSet];
					var points = pointSet.points;
					console.log("There are %1 places in '%2'".replace("%1", points.length).replace("%2", pointSet.name));
					elPointSetSelector.append($("<option/>", {
						text: testbed.name + " / " + floor.name + " / " + pointSet.name,
						value: [iTestbed, iFloor, iPointSet].join(","),
					}));
				}
			}
		}
		elPointSetSelector.selectmenu('refresh', true);
		elPointSetSelector.on('change', function() {
			var psId = getCurrentPointSetId(); // [testbedIndex, floorIndex, pointSetIndex]
			console.log("Selected: testbed " + psId[0] + ", floor " + psId[1] + ", pointSet " + psId[2]);
			showPointSet(psId);
		});
		//elPointSetSelector.trigger("change");
	}

	/**
	 * Activates the given point set on the map.
	 */
	var showPointSet = function(pointSetId) {
		var testbed = config.allTestbeds[pointSetId[0]];
		var floor = testbed.floors[pointSetId[1]];
		var iPointSet = pointSetId[2];
		var pointSet = floor.pointSets[iPointSet];
		var points = pointSet.points;

		updateWeather(pointSet.overview.centerLatLng);

		// remove previous, if any
		if (activeOverlays.mapLayer) {
			leafletMap.removeLayer(activeOverlays.mapLayer);
			activeOverlays.mapLayer = null;
		}
		if (activeOverlays.pointSetLayer) {
			leafletMap.removeLayer(activeOverlays.pointSetLayer);
			activeOverlays.pointSetLayer = null;
		}
		
		activeOverlays.mapLayer = L.imageOverlay(FLOORMAP_PATH + floor.map.image, floor.map.bounds);
		leafletMap.addLayer(activeOverlays.mapLayer);

		var geoJsonPoints = points.map(function (item) {
			return {
				"type": "Feature",
				"properties": {
					"crs": item.crs,
					"coords": item.coords,
					"comment": item.comment,
				},
				"geometry": {
					"type": "Point",
					// geoJSON uses lon/lat order, but we use lat/lon, so swap them
					"coordinates": [item.coords[1], item.coords[0]],
				}
			};
		});
		activeOverlays.pointSetLayer = L.geoJson(geoJsonPoints, {
			style: {"color": "#ff7800",},
			onEachFeature: function(feature, layer) {
				if (feature.properties) {
					var pointParams = feature.properties;

					var popup = L.popup({
						"autoPan": true,
						"closeButton": false,
						"closeOnClick": true,
					});
					var popupContent = "";
					if (!!pointParams.crs) 
						popupContent += "<p><strong>CRS:</strong> [" + pointParams.crs.join(',') + "]";
					if (!!pointParams.coords)
						popupContent += "<p><strong>Coords:</strong> [" + pointParams.coords[0].toFixed(6) + ", " + pointParams.coords[1].toFixed(6) + "]";
					if (!!pointParams.comment)
						popupContent += "<p><strong>Comment:</strong> " + pointParams.comment;
					popup.setContent(popupContent);
					layer.bindPopup(popup);

					layer.on('click', function(e) {
						selectedPoint = feature.properties;
						console.log("popup");
						// User's body partially shields RF, so the antenna can be considered
						// somewhat directional. For now this direction is not taken into account,
						// so it is fixed to "OMNI". 
						//
						// The direction is where the antenna has least shielding from.
						// For example, the user looks North, SDR in the backpack => direction is South.
						// In contrast, if the SDR is in front of the user, direction is North.
						selectedPoint.direction = 'OMNI';  // TODO add a UI control for this.
					});
				}
			},
			pointToLayer: function (feature, latlng) {
				return L.marker(latlng, {icon: pointMarkerIcon});
			},
			// Custom markers are drawn below the floorplan layer - reasons unknown...
			/*pointToLayer: function(feature, latlng) {
				return L.circleMarker(latlng, {
					radius: 10,
					fillColor: "#3f0",
					fillOpacity: 0.5,
					color: "green",
					opacity: 1,
				});
			},*/
		});
		leafletMap.addLayer(activeOverlays.pointSetLayer);
		activeOverlays.pointSetLayer.bringToFront();
		leafletMap.setView(pointSet.overview.centerLatLng, pointSet.overview.zoom);
	}; //showPointSet

	/** 
	 * Returns ID of the currently active point set.
	 * The ID is an array: [testbedIndex, floorIndex, pointSetIndex]
	 */
	var getCurrentPointSetId = function() {
		var tokens = elPointSetSelector.val().split(',');
		var iTestbed = tokens[0];
		var iFloor = tokens[1];
		var iPointSet = tokens[2];
		return [iTestbed, iFloor, iPointSet];
	};

	/** 
	 * Returns a human-readable names composing the current point set. 
	 * It is an array of [testbed_name, floor_name, pointset_name]
	 */
	var getCurrentPointSetParts = function() {
		var psId = getCurrentPointSetId();
		var iTestbed = psId[0];
		var iFloor = psId[1];
		var iPointSet = psId[2];
		var testbed = config.allTestbeds[iTestbed];
		var floor = testbed.floors[iFloor];
		return [testbed.name, floor.name, floor.pointSets[iPointSet].name];
	};

	/**
	 * Reloads current weather for the given coordinates
	 */
	var updateWeather = function(latlng) {
		elWeatherLoader.show();
		var apiUrl = WEATHER_API_URL 
						+ latlng.lat + "," + latlng.lng
						+ "?units=si"
						+ "&exclude=minutely,hourly,daily,alerts,flags"
						+ "&callback=?";
		$.getJSON(apiUrl)
			.done(function(data) {
				elSessionButton.show();
				weather = data.currently;
				elWeatherLoader.hide();
				console.log(weather);
				elWeatherSummary.text(
					(weather.summary || "(summary undefined)")
					+ ", " +
					(weather.temperature || "(undefined)") + "°C");
			});
	}

	/**
	 * Returns an object with the current session parameters
	 */
	var getSessionParams = function() {
		var data = {
			'comments': elComments.val() || "",
			'pointSetParts': getCurrentPointSetParts(),
			'pointSetId': getCurrentPointSetId(),
			'crowdActivity': elCrowdActivity.val(),
			'bands': [],
			'weather': weather,
		};
		if (!!elAcquireFM.prop("checked"))
			data.bands.push("fm");
		if (!!elAcquireGSM900down.prop("checked"))
			data.bands.push("egsm900d");
		if (!!elAcquireGSM1800down.prop("checked"))
			data.bands.push("gsm1800d");
		if (!!elAcquireDVB.prop("checked"))
			data.bands.push("dvb");
		if (!!elAcquireWiFi24.prop("checked"))
			data.bands.push("wifi24");
		if (!!elAcquireWiFi50.prop("checked"))
			data.bands.push("wifi50");

		console.log(data);
		return data;
	}

	/**
	 * Starts new measurement session or stops the existing one
	 */
	var toggleSession = function(){
		if (!sessionId) {
			// no session yet, let's start one
			elPointSetSelector.attr('disabled',true);
			//TODO disable all acquisition params, not just pointset
			$.ajax(SERVER_URL + "/startSession", {
				'type': 'POST',
				'data': getSessionParams(),
				'cache': false,
				'timeout': 1000,
				'success': function(data, textStatus, jqXHR) {
					console.log(data);
					data = JSON.parse(data);
					log("Session started [" + data.sessionId + "]");
					sessionId = data.sessionId;
					elSessionButton.text("Stop Session");
					elGetSampleButton.show();
					elStopSamplingButton.show();
				},
				'error': function(jqXHR, textStatus, errorThrown) {
					log("Session start FAILED. " + textStatus);
					sessionId = undefined;
					elGetSampleButton.hide();
					elStopSamplingButton.hide();
				}
			});
		} else {
			// session in progress, stop it
			log("Session finished.");
			sessionId = undefined;
			elSessionButton.text("Start Session");
			elPointSetSelector.attr('disabled',false);
			if (samplingInProgress)
				stopSampling();
			elGetSampleButton.hide();
			elStopSamplingButton.hide();
		}
	}

	/**
	 * Sends the server a command to collect RF sample for the given point
	 */
	function startSampling(point) {
		if (!sessionId) {
			log("No session started. Please start a session first.");
			return;
		}
		if (!selectedPoint) {
			log("No point selected. Please select a marker on the map first.");
			return;
		}
		console.log("selectedPoint", selectedPoint);
		updateSamplingState(true);
		$.ajax(SERVER_URL + "/startSampling", {
			'type': 'POST',
			'data': {
				'pointParams': selectedPoint,
			},
			'cache': false,
			'timeout': 3000,
			'success': function(data, textStatus, jqXHR) {
				console.log(data);
				// updateSamplingState(true);
				// the request initiates sampling asynchronously,
				// so sampling has just started. Updates are sent via socket.io
			},
			'error': function(jqXHR, textStatus, errorThrown) {
				log("Sampling FAILED. " + textStatus);
				updateSamplingState(false);
			}
		});
	}

	/**
	 * Sends the server a command to interrupt ongoing RF sample acqusition
	 */
	function stopSampling() {
		if (!samplingInProgress) {
			console.log("Error: trying to stop inactive sampling");
			return;
		}
		$.ajax(SERVER_URL + "/stopSampling", {
			'type': 'POST',
			'data': {},
			'cache': false,
			'timeout': 3000,
			'success': function(data, textStatus, jqXHR) {
				console.log(data);
				updateSamplingState(false);
			},
			'error': function(jqXHR, textStatus, errorThrown) {
				if (textStatus == '400') {
					// sampling was already stopped, so let's consider it a success
					updateSamplingState(false);
				} else {
					// connection problem? the request was probably not received
					log("Stop sampling FAILED: " + textStatus);
				}
			}
		});
	}

	/** Saves control states in localStorage */
	function saveControls() {
		localStorage["pointSetIndex"] = elPointSetSelector.val();
		localStorage["comments"] = elComments.val();
		localStorage["crowdActivity"] = elCrowdActivity.val();
		localStorage["acquire.fm"] = elAcquireFM.prop("checked");
		localStorage["acquire.dvb"] = elAcquireDVB.prop("checked");
		localStorage["acquire.wifi24"] = elAcquireWiFi24.prop("checked");
		localStorage["acquire.wifi50"] = elAcquireWiFi50.prop("checked");
		localStorage["acquire.gsm900d"] = elAcquireGSM900down.prop("checked");
		localStorage["acquire.gsm1800d"] = elAcquireGSM1800down.prop("checked");
	}
	/** Loads control states from localStorage */
	function loadControls() {
		elPointSetSelector.val(localStorage["pointSetIndex"] || "0,0,0").selectmenu('refresh');
		// for some reason, the above does not fire the change event, so do this manually
		elPointSetSelector.trigger("change"); 

		elCrowdActivity.val(localStorage["crowdActivity"] || 0).selectmenu('refresh');
		elComments.text(localStorage["comments"] || "");
		elAcquireFM.prop("checked",
			(localStorage["acquire.fm"] === "true") || false).checkboxradio('refresh');
		elAcquireDVB.prop("checked", 
			(localStorage["acquire.dvb"] === "true") || false).checkboxradio('refresh');
		elAcquireWiFi24.prop("checked", 
			(localStorage["acquire.wifi24"] === "true") || false).checkboxradio('refresh');
		elAcquireWiFi50.prop("checked", 
			(localStorage["acquire.wifi50"] === "true") || false).checkboxradio('refresh');
		elAcquireGSM900down.prop("checked",
			(localStorage["acquire.gsm900d"] === "true") || false).checkboxradio('refresh');
		elAcquireGSM1800down.prop("checked",
			(localStorage["acquire.gsm1800d"] === "true") || false).checkboxradio('refresh');
	}

	/** Updates controls and flags when sampling process becomes active/inactive */
	function updateSamplingState(active) {
		samplingInProgress = active;
		elGetSampleButton.attr("disabled", samplingInProgress);
		elStopSamplingButton.attr("disabled", !samplingInProgress);
	}

	/** Flashes UI with a bright color to attract user's attention */
	function uiHighlight() {
		elHighlightOverlay.show().fadeOut(1000);
	}

	// Init stuff
	elSessionButton.click(toggleSession);
	elGetSampleButton.click(startSampling);
	elStopSamplingButton.click(stopSampling);
	socket.on('child message', function(msg) {
		log(msg);
	});
	socket.on('sampling error', function(msg) {
		// fatal sampling process error
		log(msg);
		updateSamplingState(false);
		uiHighlight();
	});
	socket.on('sampling cancelled', function() {
		log("Sampling cancelled, but the already recorded files remain.");
		updateSamplingState(false);
		uiHighlight();
	})
	socket.on('sample ready', function() {
		updateSamplingState(false);
		log("*** Ready for next sample ***");
		uiHighlight();
        playBeep();
	});

	initMap(config.allTestbeds);

	$(window).unload(saveControls); // save selected choices on page close
	loadControls();
	updateSamplingState(false);
};